@@ -44,7 +44,15 @@
 #include "wayland-private.h"
 #include "wayland-os.h"
 
-#define DIV_ROUNDUP(n, a) ( ((n) + ((a) - 1)) / (a) )
+static inline uint32_t
+div_roundup(uint32_t n, size_t a)
+{
+	/* The cast to uint64_t is necessary to prevent overflow when rounding
+	 * values close to UINT32_MAX. After the division it is again safe to
+	 * cast back to uint32_t.
+	 */
+	return (uint32_t) (((uint64_t) n + (a - 1)) / a);
+}
 
 struct wl_buffer {
 	char data[4096];
@@ -308,8 +316,10 @@ wl_connection_flush(struct wl_connection *connection)
 				      MSG_NOSIGNAL | MSG_DONTWAIT);
 		} while (len == -1 && errno == EINTR);
 
-		if (len == -1)
+		if (len == -1){
+			close_fds(&connection->fds_out, MAX_FDS_OUT);
 			return -1;
+		}
 
 		close_fds(&connection->fds_out, MAX_FDS_OUT);
 
@@ -678,7 +688,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			struct wl_map *objects,
 			const struct wl_message *message)
 {
-	uint32_t *p, *next, *end, length, id;
+	uint32_t *p, *next, *end, length, length_in_u32, id;
 	int fd;
 	char *s;
 	int i, count, num_arrays;
@@ -734,8 +744,8 @@ wl_connection_demarshal(struct wl_connection *connection,
 				break;
 			}
 
-			next = p + DIV_ROUNDUP(length, sizeof *p);
-			if (next > end) {
+			length_in_u32 = div_roundup(length, sizeof *p);
+			if ((uint32_t) (end - p) < length_in_u32) {
 				wl_log("message too short, "
 				       "object (%d), message %s(%s)\n",
 				       closure->sender_id, message->name,
@@ -743,6 +753,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 				errno = EINVAL;
 				goto err;
 			}
+			next = p + length_in_u32;
 
 			s = (char *) p;
 
@@ -793,8 +804,8 @@ wl_connection_demarshal(struct wl_connection *connection,
 		case 'a':
 			length = *p++;
 
-			next = p + DIV_ROUNDUP(length, sizeof *p);
-			if (next > end) {
+			length_in_u32 = div_roundup(length, sizeof *p);
+			if ((uint32_t) (end - p) < length_in_u32) {
 				wl_log("message too short, "
 				       "object (%d), message %s(%s)\n",
 				       closure->sender_id, message->name,
@@ -802,6 +813,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 				errno = EINVAL;
 				goto err;
 			}
+			next = p + length_in_u32;
 
 			array_extra->size = length;
 			array_extra->alloc = 0;
@@ -1068,7 +1080,7 @@ buffer_size_for_closure(struct wl_closure *closure)
 			}
 
 			size = strlen(closure->args[i].s) + 1;
-			buffer_size += 1 + DIV_ROUNDUP(size, sizeof(uint32_t));
+			buffer_size += 1 + div_roundup(size, sizeof(uint32_t));
 			break;
 		case 'a':
 			if (closure->args[i].a == NULL) {
@@ -1077,7 +1089,7 @@ buffer_size_for_closure(struct wl_closure *closure)
 			}
 
 			size = closure->args[i].a->size;
-			buffer_size += (1 + DIV_ROUNDUP(size, sizeof(uint32_t)));
+			buffer_size += (1 + div_roundup(size, sizeof(uint32_t)));
 			break;
 		default:
 			break;
@@ -1139,11 +1151,11 @@ serialize_closure(struct wl_closure *closure, uint32_t *buffer,
 			size = strlen(closure->args[i].s) + 1;
 			*p++ = size;
 
-			if (p + DIV_ROUNDUP(size, sizeof *p) > end)
+			if (p + div_roundup(size, sizeof *p) > end)
 				goto overflow;
 
 			memcpy(p, closure->args[i].s, size);
-			p += DIV_ROUNDUP(size, sizeof *p);
+			p += div_roundup(size, sizeof *p);
 			break;
 		case 'a':
 			if (closure->args[i].a == NULL) {
@@ -1154,11 +1166,11 @@ serialize_closure(struct wl_closure *closure, uint32_t *buffer,
 			size = closure->args[i].a->size;
 			*p++ = size;
 
-			if (p + DIV_ROUNDUP(size, sizeof *p) > end)
+			if (p + div_roundup(size, sizeof *p) > end)
 				goto overflow;
 
 			memcpy(p, closure->args[i].a->data, size);
-			p += DIV_ROUNDUP(size, sizeof *p);
+			p += div_roundup(size, sizeof *p);
 			break;
 		default:
 			break;
