@@ -34,6 +34,7 @@
 #include <errno.h>
 #include <string.h>
 #include <unistd.h>
+#include <time.h>
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <dlfcn.h>
@@ -137,6 +138,23 @@ struct wl_protocol_logger {
 };
 
 static int debug_server = 0;
+static int debug_server_print_pid = 0;
+#define BUF_SIZE 1024
+
+void getNameByPid(pid_t pid, char *task_name) {
+    char proc_pid_path[BUF_SIZE];
+    char buf[BUF_SIZE];
+
+    sprintf(proc_pid_path, "/proc/%d/status", pid);
+    FILE* fp = fopen(proc_pid_path, "r");
+    if(NULL != fp) {
+        if(fgets(buf, BUF_SIZE-1, fp) == NULL) {
+            fclose(fp);
+        }
+        fclose(fp);
+        sscanf(buf, "%*s %s", task_name);
+    }
+}
 
 static void
 log_closure(struct wl_resource *resource,
@@ -147,8 +165,22 @@ log_closure(struct wl_resource *resource,
 	struct wl_protocol_logger *protocol_logger;
 	struct wl_protocol_logger_message message;
 
-	if (debug_server)
+	if (debug_server) {
+		if (debug_server_print_pid) {
+			time_t tt = time(NULL);
+			struct tm* t= localtime(&tt);
+			pid_t client_pid;
+			char task_name[50];
+			wl_client_get_credentials(resource->client, &client_pid, NULL, NULL);
+			getNameByPid(client_pid,task_name);
+			if (send) {
+				fprintf(stderr, "%d:%d:%d server[%d] send event to client[%d][%s]\n", t->tm_hour, t->tm_min, t->tm_sec, getpid(), client_pid, task_name);
+			} else {
+				fprintf(stderr, "%d:%d:%d server[%d] receive request from client[%d][%s]\n", t->tm_hour, t->tm_min, t->tm_sec, getpid(), client_pid, task_name);
+			}
+		}
 		wl_closure_print(closure, object, send);
+	}
 
 	if (!wl_list_empty(&display->protocol_loggers)) {
 		message.resource = resource;
@@ -353,7 +385,7 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
 		}
 	}
 
-	while ((size_t) len >= sizeof p) {
+	while (len >= 0 && (size_t) len >= sizeof p) {
 		wl_connection_copy(connection, p, sizeof p);
 		opcode = p[1] & 0xffff;
 		size = p[1] >> 16;
@@ -682,7 +714,7 @@ destroy_resource(void *element, void *data, uint32_t flags)
 	/* Don't emit the new signal for deprecated resources, as that would
 	 * access memory outside the bounds of the deprecated struct */
 	if (!resource_is_deprecated(resource))
-		wl_priv_signal_emit(&resource->destroy_signal, resource);
+		wl_priv_signal_final_emit(&resource->destroy_signal, resource);
 
 	if (resource->destroy)
 		resource->destroy(resource);
@@ -841,7 +873,7 @@ wl_client_destroy(struct wl_client *client)
 {
 	uint32_t serial = 0;
 
-	wl_priv_signal_emit(&client->destroy_signal, client);
+	wl_priv_signal_final_emit(&client->destroy_signal, client);
 
 	wl_client_flush(client);
 	wl_map_for_each(&client->objects, destroy_resource, &serial);
@@ -1002,11 +1034,16 @@ wl_display_create(void)
 {
 	struct wl_display *display;
 	const char *debug;
+	const char *debug_with_pid;
 
 	debug = getenv("WAYLAND_DEBUG");
 	if (debug && (strstr(debug, "server") || strstr(debug, "1")))
 		debug_server = 1;
 
+	debug_with_pid = getenv("WAYLAND_DEBUG_PID");
+	if (debug_with_pid && strstr(debug_with_pid, "1"))
+		debug_server_print_pid = 1;
+
 	display = malloc(sizeof *display);
 	if (display == NULL)
 		return NULL;
@@ -1089,7 +1126,7 @@ wl_display_destroy(struct wl_display *display)
 	struct wl_socket *s, *next;
 	struct wl_global *global, *gnext;
 
-	wl_priv_signal_emit(&display->destroy_signal, display);
+	wl_priv_signal_final_emit(&display->destroy_signal, display);
 
 	wl_list_for_each_safe(s, next, &display->socket_list, link) {
 		wl_socket_destroy(s);
@@ -2025,6 +2062,46 @@ wl_priv_signal_emit(struct wl_priv_signal *signal, void *data)
 	}
 }
 
+/** Emit the signal for the last time, calling all the installed listeners
+ *
+ * Iterate over all the listeners added to this \a signal and call
+ * their \a notify function pointer, passing on the given \a data.
+ * Removing or adding a listener from within wl_priv_signal_emit()
+ * is safe, as is freeing the structure containing the listener.
+ *
+ * A large body of external code assumes it's ok to free a destruction
+ * listener without removing that listener from the list.  Mixing code
+ * that acts like this and code that doesn't will result in list
+ * corruption.
+ *
+ * We resolve this by removing each item from the list and isolating it
+ * in another list.  We discard it completely after firing the notifier.
+ * This should allow interoperability between code that unlinks its
+ * destruction listeners and code that just frees structures they're in.
+ *
+ */
+void
+wl_priv_signal_final_emit(struct wl_priv_signal *signal, void *data)
+{
+	struct wl_listener *l;
+	struct wl_list *pos;
+
+	/* During a destructor notifier isolate every list item before
+	 * notifying.  This renders harmless the long standing misuse
+	 * of freeing listeners without removing them, but allows
+	 * callers that do choose to remove them to interoperate with
+	 * ones that don't. */
+	while (!wl_list_empty(&signal->listener_list)) {
+		pos = signal->listener_list.next;
+		l = wl_container_of(pos, l, link);
+
+		wl_list_remove(pos);
+		wl_list_init(pos);
+
+		l->notify(l, data);
+	}
+}
+
 /** \endcond INTERNAL */
 
 /** \cond */ /* Deprecated functions below. */
